@@ -12,89 +12,124 @@
 
 #![cfg_attr(feature = "aggressive_lint", deny(warnings))]
 #![recursion_limit = "256"]
-extern crate regex;
-
-use std::marker::Unpin;
-use std::path::Path;
-
-use crypto::digest::Digest;
-use crypto::md5::Md5;
-
-use futures::{stream, Future, TryStream, TryStreamExt};
-
-use hyper::client::connect::HttpConnector;
-use log::{info, warn};
-use log_derive::logfn;
-use tokio::task;
 
 #[cfg(feature = "blocking")]
 pub mod blocking;
 pub mod errors;
-#[cfg(feature = "http_server")]
-pub mod http_server;
 pub mod rusoto;
 
-#[cfg(feature = "compression")]
 pub(crate) mod compression;
+pub(crate) mod tempfile;
 pub(crate) mod types;
 
 mod config;
 mod ops;
 mod retry;
 
-/// Internal module used to call out operations that may block.
-mod bio {
-    pub(super) use std::fs;
-    pub(super) use std::fs::File;
-    pub(super) use std::io::prelude::*;
-    pub(super) use std::io::BufReader;
-}
+use std::marker::Unpin;
+use std::path::Path;
 
-pub use crate::errors::{Error, Result};
+use crypto::{digest::Digest, md5::Md5};
+use futures::{stream, TryStream, TryStreamExt};
+use hyper::client::connect::HttpConnector;
+use log::{info, warn};
+use log_derive::logfn;
+use tokio::{
+    fs,
+    io::{AsyncRead, AsyncReadExt, BufReader},
+};
 
+pub use crate::config::Config;
 use crate::retry::handle_dispatch_error;
 use crate::rusoto::*;
-
-use crate::config::Config;
 use crate::types::S3Listing;
 
-pub use crate::types::{ObjectInfo, S3ListingItem, S3Object, S3PathParam};
+pub use errors::{Error, Result};
+pub use ops::{
+    copy::copy,
+    download::{download, download_streaming, download_with_transparent_decompression},
+    sync::{sync, GlobFilter},
+    upload::{upload, upload_compressed, upload_from_reader, PendingUpload},
+};
+pub use rusoto::HeadObjectInfo;
+pub use types::{S3ListingItem, S3Object, S3PathParam};
 
-pub use ops::download::download;
+pub const FILTER_EMPTY: Option<&[GlobFilter]> = None;
 
-#[cfg(feature = "compression")]
-pub use ops::download::download_decompressed;
-#[cfg(feature = "compression")]
-pub use ops::upload::upload_compressed;
+pub async fn compute_etag(path: impl AsRef<Path>) -> Result<String> {
+    async fn inner(path: &Path) -> Result<String> {
+        let f = match fs::File::open(path).await {
+            Ok(f) => f,
+            Err(e) if e.kind() == std::io::ErrorKind::NotFound => {
+                return Err(Error::ETagNotPresent);
+            }
+            Err(e) => {
+                return Err(e.into());
+            }
+        };
+        let file_size = f.metadata().await?.len();
+        compute_etag_from_reader(f, file_size).await
+    }
+    inner(path.as_ref()).await
+}
 
-#[cfg(feature = "http_server")]
-pub(crate) use ops::download::download_streaming;
-
-#[cfg(feature = "cli")]
-pub use ops::upload::setup_upload_termination_handler;
-pub use ops::upload::{abort_upload, upload, upload_from_reader};
-
-pub use ops::sync::sync;
-
-pub use ops::copy::copy;
-
-pub const INCLUDE_EMPTY: Option<&[&str]> = None;
-pub const EXCLUDE_EMPTY: Option<&[&str]> = None;
-
-const EXPECT_SPAWN_BLOCKING: &str = "spawned task failed";
+pub async fn compute_etag_from_reader<T>(reader: T, length: u64) -> Result<String>
+where
+    T: AsyncRead + Unpin + Send + 'static,
+{
+    let mut reader = BufReader::new(reader);
+    let mut hash = Md5::new();
+    let mut digests: Vec<[u8; 16]> = vec![];
+    let mut remaining = length;
+    let upload_part_size = Config::global().upload_part_size();
+    while remaining != 0 {
+        let upload_part_size: usize = (if remaining >= upload_part_size {
+            upload_part_size
+        } else {
+            remaining
+        }) as usize;
+        hash.reset();
+        let mut blob = vec![0u8; upload_part_size];
+        reader.read_exact(&mut blob).await?;
+        hash.input(&blob);
+        let mut hash_bytes = [0u8; 16];
+        hash.result(&mut hash_bytes);
+        digests.push(hash_bytes);
+        remaining -= upload_part_size as u64;
+    }
+    if digests.is_empty() {
+        let mut hash_bytes = [0u8; 16];
+        hash.result(&mut hash_bytes);
+        let hex_digest = hex::encode(hash_bytes);
+        Ok(format!("\"{}\"", hex_digest))
+    } else if digests.len() == 1 && length < upload_part_size {
+        let hex_digest = hex::encode(digests[0]);
+        Ok(format!("\"{}\"", hex_digest))
+    } else {
+        let count = digests.len();
+        let mut etag_hash = Md5::new();
+        for digest_bytes in digests {
+            etag_hash.input(&digest_bytes);
+        }
+        let mut final_hash = [0u8; 16];
+        etag_hash.result(&mut final_hash);
+        let hex_digest = hex::encode(final_hash);
+        Ok(format!("\"{}-{}\"", hex_digest, count))
+    }
+}
 
 #[logfn(err = "ERROR")]
 pub async fn head_object<T>(
     s3: &T,
     bucket: impl AsRef<str>,
     key: impl AsRef<str>,
-) -> Result<Option<ObjectInfo>>
+) -> Result<Option<HeadObjectInfo>>
 where
     T: S3 + Send,
 {
     let (bucket, key) = (bucket.as_ref(), key.as_ref());
     info!("head-object: bucket={}, key={}", bucket, key);
-    head_object_request(s3, bucket, key).await
+    head_object_request(s3, bucket, key, None).await
 }
 
 #[logfn(err = "ERROR")]
@@ -232,165 +267,6 @@ where
     ))
 }
 
-pub async fn compute_etag(path: impl AsRef<Path>) -> Result<String> {
-    let path = path.as_ref();
-    if !path.exists() {
-        return Err(Error::ETagNotPresent);
-    }
-    let f = bio::File::open(path)?;
-    let stat = bio::fs::metadata(&path)?;
-    let file_size = stat.len();
-    compute_etag_from_reader(f, file_size).await
-}
-
-pub fn compute_etag_from_reader<T>(reader: T, length: u64) -> impl Future<Output = Result<String>>
-where
-    T: bio::Read + Send + 'static,
-{
-    use bio::*;
-    async move {
-        task::spawn_blocking(move || {
-            let mut reader = BufReader::new(reader);
-            let mut hash = Md5::new();
-            let mut digests: Vec<[u8; 16]> = vec![];
-            let mut remaining = length;
-            let upload_part_size = Config::global().upload_part_size();
-            while remaining != 0 {
-                let upload_part_size: usize = (if remaining >= upload_part_size {
-                    upload_part_size
-                } else {
-                    remaining
-                }) as usize;
-                hash.reset();
-                let mut blob = vec![0u8; upload_part_size];
-                reader.read_exact(&mut blob)?;
-                hash.input(&blob);
-                let mut hash_bytes = [0u8; 16];
-                hash.result(&mut hash_bytes);
-                digests.push(hash_bytes);
-                remaining -= upload_part_size as u64;
-            }
-            if digests.is_empty() {
-                let mut hash_bytes = [0u8; 16];
-                hash.result(&mut hash_bytes);
-                let hex_digest = hex::encode(hash_bytes);
-                Ok(format!("\"{}\"", hex_digest))
-            } else if digests.len() == 1 && length < upload_part_size {
-                let hex_digest = hex::encode(digests[0]);
-                Ok(format!("\"{}\"", hex_digest))
-            } else {
-                let count = digests.len();
-                let mut etag_hash = Md5::new();
-                for digest_bytes in digests {
-                    etag_hash.input(&digest_bytes);
-                }
-                let mut final_hash = [0u8; 16];
-                etag_hash.result(&mut final_hash);
-                let hex_digest = hex::encode(final_hash);
-                Ok(format!("\"{}-{}\"", hex_digest, count))
-            }
-        })
-        .await
-        .expect(EXPECT_SPAWN_BLOCKING)
-    }
-}
-
-fn process_head_obj_resp(hoo: HeadObjectOutput) -> Result<Option<ObjectInfo>> {
-    if let Some(true) = hoo.delete_marker {
-        return Ok(None);
-    }
-
-    let e_tag = if let Some(e_tag) = hoo.e_tag {
-        e_tag
-    } else {
-        return Err(Error::HeadObjectUnexpected(format!(
-            "no e_tag found: {:?}",
-            hoo
-        )));
-    };
-
-    let last_modified: String = if let Some(last_modified) = hoo.last_modified {
-        last_modified
-    } else {
-        return Err(Error::HeadObjectUnexpected("no last_modified found".into()));
-    };
-
-    let last_modified: chrono::DateTime<chrono::Utc> =
-        match chrono::DateTime::parse_from_rfc2822(&last_modified) {
-            Ok(last_modified) => last_modified.into(),
-            Err(err) => {
-                return Err(Error::HeadObjectFailedParseError(err));
-            }
-        };
-
-    let size = if let Some(content_length) = hoo.content_length {
-        content_length
-    } else {
-        return Err(Error::HeadObjectUnexpected(
-            "no content_length found".into(),
-        ));
-    };
-
-    let metadata = if let Some(metadata) = hoo.metadata {
-        metadata
-    } else {
-        return Err(Error::HeadObjectUnexpected("no metadata found".into()));
-    };
-
-    Ok(Some(ObjectInfo {
-        e_tag,
-        size,
-        last_modified,
-        metadata,
-    }))
-}
-
-#[logfn(err = "ERROR")]
-async fn head_object_request<T>(
-    s3: &T,
-    bucket: impl AsRef<str>,
-    key: impl AsRef<str>,
-) -> Result<Option<ObjectInfo>>
-where
-    T: S3,
-{
-    let (bucket, key) = (bucket.as_ref(), key.as_ref());
-    let mut res = Some(
-        handle_dispatch_error(|| async {
-            let hor = HeadObjectRequest {
-                bucket: bucket.into(),
-                key: key.into(),
-                ..Default::default()
-            };
-
-            s3.head_object(hor).await
-        })
-        .await,
-    );
-
-    match res.as_mut() {
-        Some(Ok(_)) => {
-            let hoo = res.unwrap().unwrap();
-            process_head_obj_resp(hoo)
-        }
-        Some(Err(RusotoError::Unknown(e))) => {
-            if e.status == 404 {
-                Ok(None)
-            } else {
-                let err = res.unwrap().err().unwrap();
-                Err(Error::HeadObjectFailure(err))
-            }
-        }
-        Some(Err(_)) => {
-            let err = res.unwrap().err().unwrap();
-            Err(Error::HeadObjectFailure(err))
-        }
-        _ => {
-            panic!("impossible?");
-        }
-    }
-}
-
 async fn list_objects_request<T>(
     s3: &T,
     bucket: &str,
@@ -452,12 +328,18 @@ where
     Ok(listing)
 }
 
+#[cfg(feature = "rustls")]
+pub use hyper_rustls::HttpsConnector;
+
 #[cfg(feature = "rustls")]
 pub fn new_https_connector() -> HttpsConnector<HttpConnector> {
     HttpsConnector::with_webpki_roots()
 }
 
 #[cfg(feature = "nativetls")]
-pub fn new_https_connector() -> HttpsConnector<HttpConnector> {
+pub use hyper_tls::HttpsConnector;
+
+#[cfg(feature = "nativetls")]
+pub fn new_https_connector() -> hyper_tls::HttpsConnector<HttpConnector> {
     HttpsConnector::new()
 }
